/**
 * Created by TFG on 06/29/20.
 *
 * This contains common code used by validation/loading for the AWCBL operations.
 */

var fs = require('fs');
var path = require('path');
var dateFormat = require('dateformat');
var filewatcher = require('./filewatcher');

// Defines where the monitoring areas are
var loadingFileDir = process.env.MASTER_FILE || '/opt/fresenius/efs/sftp/bcdsupport/dropbox/';
var countFiles = process.env.COUNT_FILE || '/Users/will/dev/fresenius/fresenius-m3-api/components/data/counts/';

//
// ------------------- Exporting functions --------------------------------------
//
module.exports.archiveFile = archiveFile;
module.exports.loadFile = loadFile;
module.exports.loadCounts = loadCounts;
module.exports.bulkCreateCountData = bulkCreateCountData;
module.exports.readCountData = readCountData;
module.exports.formatPrice = formatPrice;
module.exports.formatDate = formatDate;
module.exports.getCatalogId = getCatalogId;
module.exports.getDefaultCatalog = getDefaultCatalog;
module.exports.getClinicId = getClinicId;
module.exports.getVendorId = getVendorId;
module.exports.getProductId = getProductId;
module.exports.getProductVendorId = getProductVendorId;
module.exports.loadClinicProfile = loadClinicProfile;
module.exports.loadMasterItems = loadMasterItems;
module.exports.writeAwcblEntryLog = writeAwcblEntryLog;

//
// ------------------- Common functions - miscellaneus - other Bulk management -----------------------------
//

// The function move the input file in the archive directory
function archiveFile (app, file, data) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');

  var archiveDir = path.join(loadingFileDir, "archive");
  var archivePath = path.join(archiveDir, '/', path.basename(file));

  fs.exists(archiveDir, function (exists) {
    if (!exists) {
      fs.mkdir(archiveDir, function (err) {
        if (err) {
          logger.error(err);
        } else {
          fs.writeFile(archivePath, data, function (err) {
            if (err) {
              logger.error(err);
            } else {
              fs.unlink(file, function (err) {
                if (err) {
                  logger.error(err);
                } else {
                  logger.debug("moved and deleted file");
                  filewatcher.resetLoadingFlag(app);   // Re-enable watching for file to load
                }
              });
            }
          });
        }
      });
    } else {
      fs.writeFile(archivePath, data, function (err) {
        if (err) {
          logger.error(err);
        } else {
          fs.unlink(file, function (err) {
            if (err) {
              logger.error(err);
            } else {
              logger.debug("moved and deleted file");
              filewatcher.resetLoadingFlag(app);   // Re-enable watching for file to load
            }
          });
        }
      });
    }
  });
}

// The function move the input file from the prepare area into the loading area
// TBD !!!!! Copied code from archiveFile for now !! Not yet in use !
function loadFile (app, file, data) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');

  var loadingDir = path.join(loadingFileDir, "");
  var loadingPath = path.join(loadingDir, '/', path.basename(file));

  fs.exists(loadingDir, function (exists) {
    if (!exists) {
      fs.mkdir(loadingDir, function (err) {
        if (err) {
          logger.error(err);
        } else {
          fs.writeFile(loadingDir, data, function (err) {
            if (err) {
              logger.error(err);
            } else {
              fs.unlink(file, function (err) {
                if (err) {
                  logger.error(err);
                } else {
                  logger.debug("moved and deleted file");
                  filewatcher.resetLoadingFlag(app);   // Re-enable watching for file to load
                }
              });
            }
          });
        }
      });
    } else {
      fs.writeFile(loadingPath, data, function (err) {
        if (err) {
          logger.error(err);
        } else {
          fs.unlink(file, function (err) {
            if (err) {
              logger.error(err);
            } else {
              logger.debug("moved and deleted file");
              filewatcher.resetLoadingFlag(app);   // Re-enable watching for file to load
            }
          });
        }
      });
    }
  });
}

function loadCounts (app) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');

  var countData = [];
  fs.readdir(countFiles, function (err, files) {
    if (err) {
      logger.error(err);
    } else {
      for (var i = 0; i < files.length; i++) {
        fs.readFile(files[i], 'utf8', (err, data) => {
          if (err) {
            logger.error(err);
          } else {
            countData.push(data);
          }
        });
      }
    }
  });

  readCountData(countData);
}

function bulkCreateCountData (app, countDataObjects) {
  var models = app.get('models');
  var promises = [];

  countDataObjects.forEach(function (countData) {
    promises.push(countData.save());
  });

  return models.sequelize.Promise.all(promises);
}

function readCountData (app, data) {
  var models = app.get('models');

  var lines = data.split(/\r?\n/);
  var counts = [];
  for (var i = 0; i < lines.length; i++) {
    var line = lines[i].split(',');
    if (line.length > 1) {
      var count = models.CountData.build({
        timestamp: line[0],
        clinicId: line[1],
        sapMaterial: line[2],
        vendorMaterial: line[3],
        description: line[4],
        count: line[5],
        countType: line[6],
        catType: line[7],
        cntInd: line[8] === 'X'
      });
      counts.push(count);
    }
  }
  bulkCreateCountData(counts);
}

function formatPrice (app, csvPrice) {
  if (!csvPrice) {
    return csvPrice;
  }

  var flPrice;
  try {
    flPrice = parseFloat(csvPrice);
    return flPrice.toFixed(2);
  } catch (ex) {
    return "0.00";
  }
}

// The function receive a date or nothing
// If no date returns null, otherwise returns the date formatted in Iso standard.
function formatDate (app, inputDate) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');

  if (inputDate.isEmpty || inputDate.length < 2) {
    logger.debug("No discontinued date present");
    return null;
  }

  var strDate = dateFormat(inputDate, "isoDate");
  logger.debug("Set discontinued date : " + strDate);
  return strDate;
}

// Function return a valid ID if the input catalog exists in the tbl_catalogs
// Returns -1 otherwise
function getCatalogId (allCatalogs, name) {
  var out = -1;
  if (allCatalogs && name) {
    for (var v = 0; v < allCatalogs.length; v++) {
      var vendor = allCatalogs[v];
      if (vendor.catalogId === name) {
        out = vendor.id;
        break;
      }
    }
  }
  return out;
}

// Function return the ID of the default catalog
// Returns nothing otherwise
// NOTE ! The separation character used is #. NOT comma
function getDefaultCatalog ( allCatalogs) {
  var strDefCatalogs = "";    // String containg the default catalog entry ready to add

  for (var vl = 0; vl < allCatalogs.length; vl++) {
    var dv = allCatalogs[vl];
    if (dv.isDefault === true) {
      strDefCatalogs += "#" + dv.id;
    }
  }

  return strDefCatalogs;
}

// Function return a valid ID if the input clinicID exists in the tbl_clinic_profile
// Returns -1 otherwise
function getClinicId (allClinics, clinId) {
  var out = -1;
  if (allClinics && clinId) {
    for (var v = 0; v < allClinics.length; v++) {
      var record = allClinics[v];
      // The clinic ID Need to be verified without the initial dot and the alphanumeric part in the end
      // Extract the number only on both sides
      var dbClinicId = record.clinicId.match(/(\d+)/);
      var inputClinicId = clinId.match(/(\d+)/);

      if (dbClinicId[0] === inputClinicId[0]) {
        out = record.id;
        break;
      }
    }
  }
  return out;
}

// Function return a valid ID if the input vendorMaterial exists in the tbl_master_item
// Returns -1 otherwise
// The function return a valid ID on the FIRST finding !
// Duplicates are ignored at this stage
function getProductId (allProducts, vendMat) {
  var out = -1;
  if (allProducts && vendMat) {
    for (var v = 0; v < allProducts.length; v++) {
      var record = allProducts[v];
      if (record.vendorMaterial === vendMat) {
        out = record.id;
        break;
      }
    }
  }
  return out;
}


// Function return a valid ID if the input vendorID exists in the tbl_vendor_list table
// Returns -1 otherwise
function getVendorId (allVendors, vendorId) {
  var out = -1;
  if (allVendors && vendorId) {
    for (var v = 0; v < allVendors.length; v++) {
      var record = allVendors[v];
      if (record.vendorId === vendorId) {
        out = record.id;
        break;
      }
    }
  }
  return out;
}


// Function return the vandor_id for the id passed
// Returns -1 otherwise
function getProductVendorId (allProducts, id) {
  var out = "Invalid";
  if (allProducts && id) {
    for (var v = 0; v < allProducts.length; v++) {
      var record = allProducts[v];
      if (record.id === id) {
        out = record.vendorId;
        break;
      }
    }
  }
  return out;
}

// Load clinic profile for validation
function loadClinicProfile (app, debug) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');
  var models = app.get('models');

  var all_clinics = [];

  models.ClinicProfile.findAll().then(function (clinics) {
    if (clinics) {
      if (debug === true) {
        logger.debug("loaded " + clinics.length + " Clinic Profiles");
      }

      all_clinics = clinics;
      return typeof all_clinics;
    } else {
      if (debug === true) {
        logger.error("No Clinic Profiles loaded");
      }
    }
  });
}

// Load master items for validation
function loadMasterItems (app, debug) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');
  var models = app.get('models');

  var all_items = [];

  models.MasterItem.findAll().then(function (items) {
    if (items) {
      if (debug === true) {
        logger.debug("loaded " + items.length + " Master Items");
      }
      all_items = items;
      return typeof all_items;
    } else {
      if (debug === true) {
        logger.error("No Master Items loaded");
      }
    }
  });
}

// Write AWCBL LOg entry record from here
function writeAwcblEntryLog (app, path, transaction_id,  status, notes) {
  return new Promise(function (resolve, reject) {
    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var models = app.get('models');

    var record = { userName: "", reportPath: "", status: "", transactionId: "", notes: "", date: 0 };
    record.date = new Date();
    record.username = "m3-api";
    record.reportPath = path;
    record.transactionId = transaction_id;
    record.status = status;
    record.notes = notes;

//    logger.debug("writeAwcblEntryLog called - record: " + JSON.stringify(record, null, 4));

    try {
      models.AwcblLog.findOrCreate({
        where: {
          id: 0
        }, defaults: {
          userId: record.username,
          reportPath: record.reportPath,
          transactionId: record.transactionId,
          status: record.status,
          notes: record.notes,
          date: record.date
        }
      }).spread(function (data, created) {
        if (!created) {
          // Assign the data received in input (record)
          data.userId = record.username;
          data.reportPath = record.reportPath;
          data.transactionId = record.transactionId;
          data.status = record.status;
          data.notes = record.notes;
          data.date = record.date;

          data.save().then(function () {
            // Operation successfull
          }).catch(function (err) {
            // Operation failed
            reject("writeAwcblEntryLog - failed to update record " + err);
          });
        } else {
          // Operation successful - new record, should have assigned the record
        }
        resolve("writeAwcblEntryLog - OK");
      }).catch(function (err) {
        // Problems
        reject("writeAwcblEntryLog - inner SQL error " + err);
      });
    } catch (err) {
      reject("writeAwcblEntryLog - outer SQL error " + err);
    }
  });
}
