/**
 * Created by AA on 7/3/20.
 *
 * This contains the validation for the AWCBL validation operations.
 */

//var models = require('./../models/index')
var Promise = require('promise');
var fs = require('fs');
var path = require('path');
var dateFormat = require('dateformat');
var awcblcommon = require('./awcblcommon');


// The file to be validated must be loaded in the directory below
var validateFileDir = process.env.MASTER_FILE || '/opt/fresenius/efs/sftp/bcdsupport/dropbox/prepare';
var reportsFileDir = process.env.MASTER_FILE || '/opt/fresenius/m3-api/components/public/reports';
var loadingFileDir = process.env.MASTER_FILE || '/opt/fresenius/efs/sftp/bcdsupport/dropbox';


function ValidateClinicRouting(app, file) {

  return new Promise(function(resolve, reject) {

    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');


    models.ClinicProfile.findAll().then(function(clinics) {
      if (clinics) {
        logger.info("loaded " + clinics.length + " clinic profiles");

        models.VendorList.findAll().then(function(vendors) {
          if (vendors) {
            logger.debug("loaded " + vendors.length + " Vendors");

            const results = [];
            const processed_data = [];

            var anyError = false;
            fs.createReadStream(validateFileDir + '/' + file)
              .pipe(csv({
                headers: false
              }))
              .on('data', (row) => {

                var errorMessage = "";
                var status = "Pass";

                // Verify number of columns per row
                errorMessage += verifyColumnCount(row, 3);

                // Verify if clinic exists via db
                errorMessage += validateDbCheckClinicId(clinics, row[0]);

                // Verify if clinic exists via db
                errorMessage += validateDbCheckVendorId(vendors, row[1]);

                // Validate clinic_id
                errorMessage += validateClinicId(row[0]);

                // Validate route
                errorMessage += validateRouting(row[2]);

                // Validate duplicate within the file
                errorMessage += validateDuplicateInFile(row, results);

                // Get vendor_display using vendor_id
                var vendor_display = getVendorDisplay(vendors, row[1]);

                if (errorMessage !== "") {
                  status = "Fail";
                }

                if (errorMessage !== "") {
                  anyError = true;
                }

                // Prepare record for report
                rowData = {
                  clinic_id: row[0],
                  vendor_id: row[1],
                  route: row[2],
                  status: status,
                  error_desc: errorMessage
                };

                // Prepare data to be loaded in the queue
                var processedRowData =
                    row[0] + "," +
                    row[1] + ","  +
                    row[2] + "," +
                    vendor_display;

                results.push(rowData);
                processed_data.push(processedRowData);
              })
              .on('end', () => {
                const csvWriter = createCsvWriter({
                  path: reportsFileDir + '/' + file,
                  header: [{
                      id: 'clinic_id',
                      title: 'Clinic_Id'
                    },
                    {
                      id: 'vendor_id',
                      title: 'Vendor_Id'
                    },
                    {
                      id: 'route',
                      title: 'Route'
                    },
                    {
                      id: 'status',
                      title: 'Status'
                    },
                    {
                      id: 'error_desc',
                      title: 'Error_Desc'
                    }
                  ]
                });

                csvWriter
                  .writeRecords(results)
                  .then(() => console.log('Report File Created.'));


                if (anyError) {
                  logger.info('CR - There are errors. Skipping Loading File.');
                  resolve("FAIL");
                } else {
                  logger.info('CSV file successfully processed');
                  var transactionId = getHash(file + Date.now());
                  processed_data.unshift(transactionId + "," + validateFileDir + '/' + file +  "," + processed_data.length)
                  resolve(processed_data);
                }
              });
          }
        });
      }
    });
  });
}


// Validate clinic profile loading file
// Expected rows in the CSV file :
// row[0]  -> clinic_id
// row[1]  -> vendor_list
// row[2]  -> mircera_visible
// row[3]  -> mircera_io
// row[3]  -> aranesp_visible
// row[3]  -> aranesp_io
// row[3]  -> epogen_visible
// row[3]  -> epogen_io
// row[3]  -> retacrit_visible
// row[3]  -> retacrit_io
// row[3]  -> countin_each
function ValidateClinicProfile(app, file) {

  return new Promise(function(resolve, reject) {
    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    var list_catalogs;

    // Need to load two tables for validate data
    // - clinic_profile
    // - catalogs

    models.ClinicProfile.findAll().then(function(clinics) {
      if (clinics) {
        logger.info("loaded " + clinics.length + " clinic profiles");

        models.CatalogList.findAll().then(function(catalogs) {
          if (catalogs) {
            logger.info("loaded " + catalogs.length + " catalogs");

            const results = [];
            const processed_data = [];

            var anyError = false;
            fs.createReadStream(validateFileDir + '/' + file)
                .pipe(csv({
                  headers: false
                }))
                .on('data', (row) => {

                  var errorMessage = "";
                  var status = "Pass";

                  // Verify number of columns per row
                  errorMessage += verifyColumnCount(row, 11);

                  // Verify if clinic exists via db
                  errorMessage += validateDbCheckClinicId(clinics, row[0]);

                  // Verify if catalogs exists via db
                  errorMessage += validateCatalogList(logger, catalogs, row[1]);

                  // Validate clinic_id
                  errorMessage += validateClinicId(row[0]);

                  // Validate yes or no values
                  for (var i = 2; i < 11; i++) {
                    errorMessage += validateYesNo(row[i]);
                  }

                  // Validate duplicate within the file
                  errorMessage += validateDuplicateInFile(row, results);

                  if (errorMessage !== "")
                    status = "Fail";

                  if (errorMessage !== "") {
                    anyError = true;
                  }

                  // Prepare the list of catalog ID to be sent to the queue
                  // Verify if clinic exists via db
                  list_catalogs = returnCatalogListAsId(logger, catalogs, row[1]);

                  // Prepare record for report
                  rowData = {
                    clinic_id: row[0],
                    vendor_list: row[1],    // Dont report the converted IDs
                    mircera_visible: row[2],
                    mircera_io: row[3],
                    aranesp_visible: row[4],
                    aranesp_io: row[5],
                    epogen_visible: row[6],
                    epogen_io: row[7],
                    retacrit_visible: row[8],
                    retacrit_io: row[9],
                    countin_each: row[10],
                    status: status,
                    error_desc: errorMessage
                  };

                  // Prepare data to be loaded in the queue
                  var processedRowData =
                      row[0] + "," +
                      list_catalogs + "," +
                      row[2] + ","  +
                      row[3] + "," +
                      row[4] + "," +
                      row[5] + "," +
                      row[6] + "," +
                      row[7] + "," +
                      row[8] + "," +
                      row[9] + "," +
                      row[10];

                  results.push(rowData);
                  processed_data.push(processedRowData);
                })
                .on('end', () => {
                  const csvWriter = createCsvWriter({
                    path: reportsFileDir + '/' + file,
                    header: [{
                      id: 'clinic_id',
                      title: 'Clinic_Id'
                    },
                      {
                        id: 'vendor_list',
                        title: 'Vendor_list'
                      },
                      {
                        id: 'mircera_visible',
                        title: 'Mircera_visible'
                      },
                      {
                        id: 'mircera_io',
                        title: 'Mircera_io'
                      },
                      {
                        id: 'aranesp_visible',
                        title: 'Aranesp_visible'
                      },
                      {
                        id: 'aranesp_io',
                        title: 'Aranesp_io'
                      },
                      {
                        id: 'epogen_visible',
                        title: 'Epogen_visible'
                      },
                      {
                        id: 'retacrit_visible',
                        title: 'Retacrit_visible'
                      },
                      {
                        id: 'retacrit_io',
                        title: 'Retacrit_io'
                      },

                      {
                        id: 'countin_each',
                        title: 'Countin_each'
                      },
                      {
                        id: 'status',
                        title: 'Status'
                      },
                      {
                        id: 'error_desc',
                        title: 'Error_Desc'
                      }
                    ]
                  });

                  csvWriter
                      .writeRecords(results)
                      .then(() => console.log('Report File Created.'));


                  if (anyError) {
                    logger.info('CP - There are errors. Skipping Loading File.');
                    resolve("FAIL");
                  } else {
                    logger.info('CSV file successfully processed');
                    var transactionId = getHash(file + Date.now());
                    processed_data.unshift(transactionId + "," + validateFileDir + '/' + file +  "," + processed_data.length);
                    resolve(processed_data);
                  }
                });
          } else {
            logger.error("Loading Catalog table failed !");
          }
        });
      } else {
        logger.error("Loading Clinic Profile table failed !");
      }
    });
  });
}

// Validate master item loading file
// Expected rows in the CSV file :
// row[0]  -> sap_material
// row[1]  -> description
// row[2]  -> cnt_ind
// row[3]  -> cnt_description
// row[4]  -> med_vendor
// row[5]  -> vendor_material
// row[6]  -> conv
// row[7]  -> count_type
// row[8]  -> alternates
// row[9]  -> unit_price
// row[10] -> vendor_id
// row[11] -> ndc
// row[12] -> onhand_inventory_ind
// row[13] -> discontinued
function ValidateMasterItem(app, file) {
  return new Promise(function(resolve, reject) {

    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    var list_catalogs;

    models.VendorList.findAll().then(function (vendors) {
      if (vendors) {
        logger.debug("loaded " + vendors.length + " Vendors");

        models.MasterItem.findAll().then(function (items) {
          if (items) {
            logger.debug("loaded " + items.length + " Vendors");

            models.CatalogList.findAll().then(function (catalogs) {
              if (catalogs) {
                logger.info("loaded " + catalogs.length + " catalogs");

                const results = [];
                const processed_data = [];
                const vendor_materials = [];

                var anyError = false;
                fs.createReadStream(validateFileDir + '/' + file)
                    .pipe(csv({
                      headers: false
                    }))
                    .on('data', (row) => {
                      vendor_materials.push(row[5]);
                    })
                    .on('end', () => {
                      fs.createReadStream(validateFileDir + '/' + file)
                          .pipe(csv({
                            headers: false
                          }))
                          .on('data', (row) => {

                            var errorMessage = "";
                            var status = "Pass";

                            // Verify number of columns per row
                            errorMessage += verifyColumnCount(row, 14);

                            // Verify if vendor exists via db
                            errorMessage += validateDbCheckVendorId(vendors, row[10]);

                            // Verify if vendor exists via db
                            errorMessage += validateValidDate(row[13]);

                            // Verify alternate
                            if (row[8] !== "") {
                              errorMessage += validateAlternate(items, vendor_materials, row[8]);
                            }

                            // Verify if catalogs exists via db
                            errorMessage += validateCatalogList(logger, catalogs, row[4]);

                            // Validate boolean entries
                            logger.debug("  ---> validateX on : " + row[2]);
                            errorMessage += validateX(row[2]);
                            logger.debug("  ---> validateX on : " + row[12]);
                            errorMessage += validateX(row[12]);

                            // Validate duplicate within the file
                            //errorMessage += validateDuplicateInFile(row, results);

                            if (errorMessage !== "") {
                              status = "Fail";
                            }

                            if (errorMessage !== "") {
                              anyError = true;
                            }

                            // Prepare the list of catalog ID to be sent to the queue
                            // Verify if clinic exists via db
                            list_catalogs = returnCatalogListAsId(logger, catalogs, row[4]);

                            rowData = {
                              sap_material: row[0],
                              description: row[1],
                              cnt_ind: row[2],
                              cnt_description: row[3],
                              med_vendor: row[4],
                              vendor_material: row[5],
                              conv: row[6],
                              count_type: row[7],
                              alternates: row[8],
                              unit_price: row[9],
                              vendor_id: row[10],
                              ndc: row[11],
                              onhand_inventory_ind: row[12],
                              discontinued: row[13],

                              status: status,
                              error_desc: errorMessage
                            };

                            // Prepare data to be loaded in the queue
                            var processedRowData =
                                row[0] + "," +
                                row[1] + "," +
                                row[2] + "," +
                                row[3] + "," +
                                list_catalogs + "," +
                                row[5] + "," +
                                row[6] + "," +
                                row[7] + "," +
                                row[8] + "," +
                                row[9] + "," +
                                row[10] + "," +
                                row[11] + "," +
                                row[12] + "," +
                                row[13];

                            results.push(rowData);
                            processed_data.push(processedRowData);
                          })
                          .on('end', () => {
                            const csvWriter = createCsvWriter({
                              path: reportsFileDir + '/' + file,
                              header: [{
                                id: 'sap_material',
                                title: 'Sap_Material'
                              },
                                {
                                  id: 'description',
                                  title: 'Description'
                                },
                                {
                                  id: 'cnt_ind',
                                  title: 'Cnt_Ind'
                                },
                                {
                                  id: 'cnt_description',
                                  title: 'Cnt_Description'
                                },
                                {
                                  id: 'med_vendor',
                                  title: 'Med_Vendor'
                                },
                                {
                                  id: 'vendor_material',
                                  title: 'Vendor_material'
                                },
                                {
                                  id: 'conv',
                                  title: 'Conv'
                                },
                                {
                                  id: 'count_type',
                                  title: 'Count_Type'
                                },
                                {
                                  id: 'alternates',
                                  title: 'Alternates'
                                },
                                {
                                  id: 'unit_price',
                                  title: 'Unit_Price'
                                },
                                {
                                  id: 'vendor_id',
                                  title: 'Vendor_Id'
                                },
                                {
                                  id: 'ndc',
                                  title: 'Ndc'
                                },
                                {
                                  id: 'onhand_inventory_ind',
                                  title: 'Onhand_inventory_ind'
                                },
                                {
                                  id: 'discontinued',
                                  title: 'Discontinued'
                                },
                                {
                                  id: 'status',
                                  title: 'Status'
                                },
                                {
                                  id: 'error_desc',
                                  title: 'Error_Desc'
                                }
                              ]
                            });

                            csvWriter
                                .writeRecords(results)
                                .then(() => console.log('Report File Created.'));


                            if (anyError) {
                              logger.info('MT - There are errors. Skipping Loading File.');
                              resolve("FAIL");
                            } else {
                              logger.info('CSV file successfully processed');
                              var transactionId = getHash(file + Date.now());
                              processed_data.unshift(transactionId + "," + validateFileDir + '/' + file +  "," + processed_data.length);
                              resolve(processed_data);
                            }
                          });
                    });
              } else {
                logger.error("Loading Catalog table failed !");
              }
            });
          } else {
            logger.error("Loading Master Items table failed !");
          }
        });
      } else {
        logger.error("Loading Vendor List table failed !");
      }
    });
  });
}

function ValidateParValue(app, file) {

  return new Promise(function(resolve, reject) {

    const csv = require('csv-parser');
    const createCsvWriter = require('csv-writer').createObjectCsvWriter;

    var winston = app.get('winston');
    var logger = winston.loggers.get('main');
    var transactionLogger = winston.loggers.get('transaction');
    var models = app.get('models');

    var vendor_id;
    var material_description;


    models.ClinicProfile.findAll().then(function(clinics) {
      if (clinics) {
        logger.info("loaded " + clinics.length + " clinic profiles");


        models.MasterItem.findAll().then(function(items) {
          if (items) {
            logger.debug("loaded " + items.length + " Vendors");

            const results = [];
            const processed_data = [];

            var anyError = false;
            fs.createReadStream(validateFileDir + '/' + file)
              .pipe(csv({
                headers: false
              }))
              .on('data', (row) => {

                var errorMessage = "";
                var status = "Pass";

                try {
                  // Verify number of columns per row
                  errorMessage += verifyColumnCount(row, 5);

                  // Verify if clinic exists via db
                  errorMessage += validateDbCheckClinicId(clinics, row[0]);

                  // Verify if clinic exists via db
                  errorMessage += validateVendorMaterial(items, row[1]);

                  // Validate clinic_id
                  errorMessage += validateClinicId(row[0]);

                  // Validate duplicate within the file
                  errorMessage += validateDuplicateInFile(row, results);

                  // Get vendor_id using vendor_material
                  vendor_id = getVendorId(items, row[1]);

                  // Get material_description using vendor_material
                  material_description = getMaterialDesc(items, row[1]);

                } catch (e) {
                  logger.error('TFG - There are errors. ');
                }
                if (errorMessage !== ""){
                  status = "Fail";
                  anyError = true;
//                  logger.error("TFG - PV found error : " + errorMessage);
                }

                // Prepare record for report
                rowData = {
                  clinic_id: row[0],
                  vendor_material: row[1],
                  qoh: row[2],
                  min_par: row[3],
                  max_par: row[4],
                  status: status,
                  error_desc: errorMessage
                };

                // Prepare data to be loaded in the queue
                var processedRowData =
                    row[0] + "," +
                    row[1] + "," +
                    row[2] + "," +
                    row[3] + "," +
                    row[4] + "," +
                    vendor_id + "," +
                    material_description;

                results.push(rowData);
                processed_data.push(processedRowData);

              })
              .on('end', () => {
                const csvWriter = createCsvWriter({
                  path: reportsFileDir + '/' + file,
                  header: [{
                      id: 'clinic_id',
                      title: 'Clinic_Id'
                    },
                    {
                      id: 'vendor_material',
                      title: 'Vendor_Material'
                    },
                    {
                      id: 'qoh',
                      title: 'Qoh'
                    },
                    {
                      id: 'min_par',
                      title: 'Min_Par'
                    },
                    {
                      id: 'max_par',
                      title: 'Max_Par'
                    },
                    {
                      id: 'status',
                      title: 'Status'
                    },
                    {
                      id: 'error_desc',
                      title: 'Error_Desc'
                    }
                  ]
                });

                csvWriter
                  .writeRecords(results)
                  .then(() => console.log('Report File Created.'));


                if (anyError) {
                  logger.info('PV - There are errors. Skipping Loading File.');
                  resolve("FAIL");
                } else {
                  logger.info('CSV file successfully processed');
                  var transactionId = getHash(file + Date.now());
                  processed_data.unshift(transactionId + "," + validateFileDir + '/' + file +  "," + processed_data.length);
                  resolve(processed_data);
                }
              });

          }
        });
      }
    });

  });
}




function validateRouting(data) {
  if (data !== 'IO' && data !== 'SAP') {
    return "Invalid Routing: " + data + ". ";
  } else {
    return "";
  }
}

function validateYesNo(data) {
  if (data !== 'y' && data !== 'n') {
    return "Invalid y/n value: " + data + ". ";
  } else {
    return "";
  }
}

function validateX(data) {
  if (data.length > 1 || (data.length === 1 && data !== 'X')) {
    return "Expected X or nothing: " + data + ". ";
  } else {
    return "";
  }
}

function validateDbCheckClinicId(clinics, clinic_id) {
  var id = 1;
  id = awcblcommon.getClinicId(clinics, clinic_id);
  if (id < 0) {
    //return "Clinic_Id: " + clinic_id + " is not present in the db. id =  " + id + ". ";
    return "Clinic_Id: " + clinic_id + " is not present in the db. ";
  } else {
    return "";
    //return "Clinic_Id: " + clinic_id + " is  present in the db.  id =  " + id + ". ";
  }
}


function validateDbCheckVendorId(vendors, vendor_id) {
  var id = 1;
  id = awcblcommon.getVendorId(vendors, vendor_id);
  if (id < 0) {
    //return "Vendor_Id: " + vendor_id + " is not present in the db.  id =  " + id + ". ";
    return "Vendor_Id: " + vendor_id + " is not present in the db. ";
  } else {
    return "";
  //return "Vendor_Id: " + vendor_id + " is  present in the db.  id =  " + id + ". ";
  }
}

function validateVendorMaterial(allMasterItems, vendMat) {
  var id = 1;
  id = awcblcommon.getProductId(allMasterItems, vendMat);
  if (id < 0) {
    //return "Vendor Material: " + vendMat + " is not present in the db.  id =  " + id + ". ";
    return "Vendor Material: " + vendMat + " is not present in the db. ";
  } else {
    return "";
    //return "Vendor Material: " + vendMat + " is  present in the db.  id =  " + id + ". ";
  }
}

function validateAlternate(allMasterItems, vendor_materials, alternate) {
  var found = false;
  for (var v = 0; v < allMasterItems.length; v++) {
    var record = allMasterItems[v];
    if (record.vendorMaterial === alternate && !record.discontinued) {
      found = true;
//      return "";
      break;
    }
  }

  if (!found) {
    for (var i = 0; i < vendor_materials.length; i++) {
      if (vendor_materials[i] === alternate) {
        found = true;
//        return "";
        break;
      }
    }

    if (found) {
      return "";
    } else {
      return "Invalid alternate: " + alternate + ". ";
    }
  }
}

function validateClinicId(data) {
  var rePattern = new RegExp(/^\.([0-9]+)(BM[0-9]*)$|(HP$)/);
  var result = rePattern.test(data);

  if (result === false) {
    return "Invalid clinic_id: " + data + ". ";
  } else {
    return "";
  }
}

// The function analyze an input string containing a list of catalogs
// and return an error, empty otherwise
function validateCatalogList(logger, catalogList, data) {
  logger.debug('-----------------------------------------------------------');
  logger.debug('validateCatalogList on : ' + data.toString());

  // Checking for the presence of the invalid characters in the input !
  if (data.toString().includes("#")) {
    return "Presence of forbidden characters (#) on " + data;
  }

  if (data.length < 2 || data === "") {
    // Empty list !!   Verify if there are default catalogs
    var strDefCatalogs = awcblcommon.getDefaultCatalog(catalogList);
    if (strDefCatalogs.length > 1) {
      // Empty list but is present at least a default catalog
      logger.debug("  --> Empty list - OK");
      return "";
    } else {
      // Empty list and NO default catalogs !! Error !
      logger.debug("  --> Empty list - ERROR");
      return "No catalogs to be assigned to the profile !";
    }
  }

  var testCatalog = data.toString().split('"').join('');
  var fileVendorNames = testCatalog.split(",");

  logger.debug('fileVendorNames array length : ' + fileVendorNames.length);

  var cnt = 0;

  if (testCatalog !== "") {
    // catalogs in the input name - verify exists and convert in index

    for (var fn = 0; fn < fileVendorNames.length; fn++) {
      logger.debug('  --> catalog name under test : ' + fileVendorNames[fn].trim());

      var id = awcblcommon.getCatalogId(catalogList, fileVendorNames[fn].trim());
      if (id >= 0) {
        logger.debug("Found catalog [" + fileVendorNames[fn].trim() + "] with index " + id);
      } else {
        return "Catalog parsing error on [" + fileVendorNames[fn] + "] ";
      }
      cnt++;
    }
  }

  return "";
}

// The function convert the catalog list in IDs
// It return a strinf containing the list of the catalog ID present in the data
// or at least a default catalog
function returnCatalogListAsId(logger, catalogList, data) {
  var returnResult = "";
  var isDefaultPresent = false;
  var strDefCatalogs = "";    // String containg the default catalog entry ready to add
  var testCatalog = data.split('"').join('');
  var fileVendorNames = testCatalog.split(",");
  var cnt = 0;

  logger.debug('returnCatalogListAsId on : ' + data.toString());

  if (!testCatalog) {
    testCatalog = "";
  }

  // Assign default catalogs
  strDefCatalogs = awcblcommon.getDefaultCatalog(catalogList);

  logger.debug('  --> default catalog ID  : ' + strDefCatalogs);
  logger.debug('  --> testCatalog line    : ' + testCatalog);

  if (testCatalog !== "") {
    // catalogs in the input name - verify exists and convert in index

    for (var fn = 0; fn < fileVendorNames.length; fn++) {
      var id = awcblcommon.getCatalogId(catalogList, fileVendorNames[fn].trim());
      if (id >= 0) {
        // Found an ID !
        if (id.toString() === strDefCatalogs.substr(1)) {
          logger.debug("Found default catalog !");
          isDefaultPresent = true;
        } else {
          logger.debug("Found catalog [" + fileVendorNames[fn].trim() + "] with index " + id);
        }

        // Don't use the comma as separator !!  Uses #
        returnResult = returnResult + ((cnt > 0) ? "#" : "") + id;
        cnt++;
      } else {
        // No valid ID found !
        // The error is handled anyway in the validation check - do noting
        logger.debug("Catalog parsing error on [" + fileVendorNames[fn] + "] ");
      }
    }

    // If the default catalog is NOT present, add it to the list
    if (isDefaultPresent === false) {
      logger.debug("Add Default catalog to the list");
      returnResult += strDefCatalogs;
    }

  } else {
    // No catalogs found - add at least the default one
    returnResult = strDefCatalogs.substr(1);
  }

  logger.debug('  --> returning result : ' + returnResult);
  return returnResult;
}


function validateValidDate(data) {
  if (data === "") {
    return "";
  } else {
    var value = isValidDate(data);
    if (value) {
      return "";
    } else {
      return "Invalid discontinued date: " + data + ". ";
    }
  }
}

function isValidDate(date) {
  try {
    var dateWrapper = new Date(date);
    return dateWrapper && Object.prototype.toString.call(dateWrapper) === "[object Date]" && !isNaN(dateWrapper);
  } catch (err) {
    return false;
  }
}

function validateDuplicateInFile(row, results) {

  var errorMessage = "";

  for (var i = 0; i < results.length; i++) {
    if (row[0] === results[i]["clinic_id"]) {
      if (row[1] === results[i]["vendor_id"]) {
        errorMessage = "Duplicate data in the file: " + row[0] + "," + row[1] + ". ";
        break;
      }
    }
  }

  return errorMessage;
}


function RemoveFileFromValidationArea(app, file) {
  var winston = app.get('winston');
  var logger = winston.loggers.get('main');
  var transactionLogger = winston.loggers.get('transaction');

  try {
    fs.unlinkSync(filePath)
    logger.info("Removed File: " + validateFileDir + '/' + file);
    //file removed
  } catch (err) {
    console.error(err)
  }
}

function verifyColumnCount(obj, expectedCount) {
  if (Object.keys(obj).length !== expectedCount)
    return "Invalid column count: " + Object.keys(obj).length + ", expected: " + expectedCount + ". ";
  else
    return "";
}

function getVendorId(allMasterItems, vendor_material) {
  for (var i = 0; i < allMasterItems.length; i++) {
    var record = allMasterItems[i];
    if (record.vendorMaterial === vendor_material) {
      return record.vendorId;
    }
  }
}

function getMaterialDesc(allMasterItems, vendor_material) {
  for (var i = 0; i < allMasterItems.length; i++) {
    var record = allMasterItems[i];
    if (record.vendorMaterial === vendor_material) {
      return record.description;
    }
  }
}

function getVendorDisplay(vendors, vendor_id) {
  for (var i = 0; i < vendors.length; i++) {
    var record = vendors[i];
    if (record.vendorId === vendor_id) {
      return record.vendorDisplay;
    }
  }
}

function getHash(input){
  var hash = 0, len = input.length;
  for (var i = 0; i < len; i++) {
    hash  = ((hash << 5) - hash) + input.charCodeAt(i);
    hash |= 0; // to 32bit integer
  }
  return hash;
}

module.exports.RemoveFileFromValidationArea = RemoveFileFromValidationArea;
module.exports.ValidateClinicRouting = ValidateClinicRouting;
module.exports.ValidateClinicProfile = ValidateClinicProfile;
module.exports.ValidateParValue = ValidateParValue;
module.exports.ValidateMasterItem = ValidateMasterItem;